# Copyright 2017, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0.
# See the LICENSE file associated with the project for terms.

from .config import \
    ConfigSourceBase, \
    DictConfigSource, \
    ChainedConfigSource, \
    YamlConfigSource, \
    CliConfigSource

from .sink import \
    BaseTemplate, \
    FileTemplate, \
    Sink, \
    FileSink, \
    SinkSupplier, \
    ExtensionFileSinkSupplier

__version__ = '0.0.1'



# Copyright 2017, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0.
# See the LICENSE file associated with the project for terms.

try:
    from collections import ChainMap
except:
    # py2 has extra dependency
    from chainmap import ChainMap

try:
    string_type = basestring
except NameError:
    string_type = str



# Copyright 2017, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0.
# See the LICENSE file associated with the project for terms.

"""
Built-in configuration sinks for interpolatr
"""

from contextlib import closing
import click
import os
import logging
import re
from collections import Iterator

from jinja2 import Environment, FileSystemLoader, contextfunction

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

TOKEN_PAT = re.compile('^\$\((.*)\)$')


def replace_regex_jinja(inp, find, replace):
    """ Jinja regex replace function """
    return re.sub(find, replace, inp)


@contextfunction
def default_finalize(ctx, thing):
    """
    Default finalizer used for interpolation with Jinja.

    Does two things:
        - coalesce false-y values to ''
        - A hacky 1-level lookup back into the context if we get something
          that looks like a key. (Assumes keys look like $(...).
    """
    if not thing:
        return ''

    m = TOKEN_PAT.match(str(thing))
    if m:
        nested = ctx.get(m.group(1))
        return nested if nested is not None else ''
    else:
        return thing


# You could override this if you want,
# but it does defaults nice things for you
def build_default_env(loader, filters=None, **kwargs):
    """
    Build jinja environment with default finalizer and some other defaults.

    :param loader: jinja loader
    :keyword filters: filters to use

    Other parameters in kwargs will be passed directly to the environment.
    """
    kwargs['loader'] = loader

    filters = filters or {'replace_regex': replace_regex_jinja}

    # Defaults
    if 'variable_start_string' not in kwargs:
        kwargs['variable_start_string'] = '$('

    if 'variable_end_string' not in kwargs:
        kwargs['variable_end_string'] = ')'

    if 'finalize' not in kwargs:
        kwargs['finalize'] = default_finalize

    env = Environment(
        **kwargs
    )

    env.filters.update(filters)

    return env


class BaseTemplate(object):
    """
    Source of jinja2 templates. Renders self (returns a generator for chunks
    of rendered content) via the `interpolate` method.

    Templates take a `sink` via the constructor. The sink should be a `Sink`
    object (see below). The `commit` method interpolates the template
    using the provided jinja2 context and writes the resulting chunks to the
    sink.
    """

    # subclass should call super() in constructor
    def __init__(self, sink):
        self.sink = sink

    def interpolate(self, context):
        """
        Interpolate template given context.

        :param context: dict-like object
        """
        t = self.get_template()
        return t.generate(context)

    def get_template(self):
        """
        Return a jinja2 template object.
        """
        raise NotImplementedError('Not implemented!')

    def commit(self, context):
        """
        Commit the interpolated template to somewhere.

        :param context: dict-like object
        """

        # Open the sink object for writing
        self.sink.open()

        with closing(self.sink):
            for chunk in self.interpolate(context):
                self.sink.write(chunk)


class FileTemplate(BaseTemplate):
    """
    A file template source.
    """
    def __init__(self, sink, env, path):
        super(FileTemplate, self).__init__(sink)

        self.env = env
        self.path = path

    def get_template(self):
        return self.env.get_template(self.path)


class Sink(object):
    """
    A sink should render a template given a context, and must `commit` it
    (presumably, to some persistent store such as a file).
    """
    def open(self):
        """
        Open the sink for writing. May be a no-op.
        """
        pass

    def close(self):
        """
        Close the sink. May be a no-op.
        """
        pass

    def write(self, chunk):
        """
        Write a chunk of interpolated data to the sink. Must be implemented.

        :param chunk: chunk of content to write
        """
        raise NotImplementedError('Not implemented!')


class FileSink(Sink):
    """
    A file sink can render templates to the given file.
    """
    def __init__(self, outpath):
        self.outpath = outpath
        self.fh = None

    def open(self):
        self.fh = open(self.outpath, 'w')

    def close(self):
        self.fh.close()
        self.fh = None

    def write(self, chunk):
        self.fh.write(chunk)


class SinkSupplier(Iterator):

    def __init__(self):
        self._iter = None

    # This is silly, but is an artifact of how these are created.
    @classmethod
    def object_type(cls):
        return 'SinkSupplier'

    # python2 wants next()
    def next(self):
        return self.__next__()

    # python3 wants __next__
    def __next__(self):
        if self._iter is None:
            self._iter = self.generate_outputs()
        return next(self._iter)

    def generate_outputs(self):
        """
        Generator to create template objects. Typical usage would be to treat
        the supplier as an iterator, commit()ing each object returned.
        """
        raise NotImplementedError('Not implemented!')


class ExtensionFileSinkSupplier(SinkSupplier):
    """
    Generator for FileSinks using an extension to find files
    """

    def __init__(self, path, extension='.interpolate', recursive=True):
        """
        Create extension sink supplier

        :param path: base template search path
        :keyword extension: extension to search for (default .interpolate)
        :keyword recursive: search recursively (default true)
        """
        super(ExtensionFileSinkSupplier, self).__init__()
        self.path = path
        self.extension = extension
        self.recursive = recursive
        self.env = build_default_env(
            FileSystemLoader(path)
        )

    def generate_outputs(self):
        for root, dirs, files in os.walk(self.path):
            plain_files = (f for f in files if os.path.isfile(
                os.path.join(root, f)))
            for filename in plain_files:
                fpath = os.path.join(root, filename)

                if os.path.isfile(fpath) and fpath.endswith(self.extension):
                    output_file = os.path.splitext(fpath)[0]
                    template_path = os.path.relpath(fpath, self.path)
                    yield FileTemplate(
                        FileSink(output_file),
                        self.env,
                        template_path
                    )
            if not self.recursive:
                break

    @classmethod
    def create(cls, target_dir=None, extension='.interpolate', recursive=True):
        # Normalize the target directory
        target_dir = os.path.abspath(target_dir)
        if not os.path.exists(target_dir):
            raise Exception('target_dir {0} does not exist'.format(target_dir))

        return cls(target_dir, extension, recursive)

    @classmethod
    def get_args(cls):
        """
        Return the array of arguments to click. For use if specified
        on the command line.
        """
        return [
            click.Option(('--recursive/--no-recursive',),
                         help='Recurse into search dir.', default=True),
            click.Option(('--extension',),
                         help='Extension of files to search for.',
                         default='.interpolate'),
            click.Option(('--target_dir',), '-t', help='Search directory.',
                         default=os.getcwd())
        ]



# Copyright 2017, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0.
# See the LICENSE file associated with the project for terms.

"""
Built-in configuration sources.

Some, like DictConfigSource or ChainedConfigSource, are mostly just
wrappers around existing python types.

Others, like the YamlConfigSource, are a little more interesting.
"""

import collections
import logging
import os
import re

import click
import yaml

from .compat import ChainMap
from .compat import string_type


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

TOKEN_PAT = re.compile('^\$\((.*)\)$')


class ConfigSourceBase(collections.Mapping):
    """
    A configuration source isn't much more than a mapping.

    It is used as the context passed to your template.

    `ConfigSourceBase` is abstract.

    To be callable from built-in command line tool, your class must implement
    two class methods:
        - get_args: return a list of parameters for click
        - create: create a ConfigSourceBase. create() should accept the
          arguments specified by get_args(). Note that a class method is used
          instead of __new__ because you may want to return a different object
          and doing that from __new__ just confuses people.
    """

    # This is silly, but is necessary for creating these things via the CLI
    @classmethod
    def object_type(cls):
        return 'ConfigSource'

    def dump(self):
        """
        Get a one-per-line view of the map.
        """
        return '\n'.join('{0}={1}'.format(k, self[k]) for k in sorted(
            self.keys())
        )

    def classy_dump(self):
        """
        Get a compactly-formatted view of the map contents, including
        the class name. Is used as the default implementation of __repr__().
        """
        return '{0}({1})'.format(
            self.__class__.__name__,
            ','.join('{0}={1}'.format(k, self[k]) for k in sorted(self.keys()))
        )

    def __repr__(self):
        return self.classy_dump()


class DictConfigSource(ConfigSourceBase):
    """
    The simplest config source.

    :param other: dictionary to use a config source
    """
    def __init__(self, other):
        self.other = other

    def __getitem__(self, key): return self.other[key]

    def __len__(self): return len(self.other)

    def __iter__(self): return iter(self.other)


class ChainedConfigSource(ConfigSourceBase, ChainMap):
    """
    A ConfigSource that isa ChainMap.

    Note: this thing used to never throw KeyError (always returned
    None). I've changed this behavior now, but can't remember the reason for
    the original version...
    """
    def __init__(self, *args):
        super(ChainedConfigSource, self).__init__(*args)

    def chain(self, mapping):
        """
        Append a mapping to the end of the chain.
        """
        self.maps.append(mapping)

    def __repr__(self):
        return '{0}({1})'.format(
            self.__class__.__name__,
            ','.join(str(i) for i in self.maps)
        )

    def __str__(self):
        return '{0}(\n    {1}\n)'.format(
            self.__class__.__name__,
            ',\n    '.join(str(i) for i in self.maps)
        )


class YamlConfigSource(ConfigSourceBase):
    """
    Configuration from special YAML file format.
    """

    def __init__(self, path=None):
        if not os.path.isfile(path):
            raise Exception("Could not find file {0}".format(path))

        # FIXME: Not clear -- should these paths be relative to this file,
        # or to the cwd?
        self.abspath = os.path.abspath(path)
        self.dirpath = os.path.dirname(self.abspath)

        conf = None
        with open(self.abspath) as f:
            conf = yaml.load(f.read())
            if not conf:
                logger.warning(
                    'Configuration file {0} seems to be empty'.format(
                        self.abspath
                    )
                )
                conf = {}

        self.settings = conf.get('settings', {})
        self.base = conf.get('base')

    def get_base(self):
        return self.base

    @classmethod
    def get_args(cls):
        """
        Return the array of arguments to click. For use if specified
        on the command line.
        """
        return [click.Option(("--path",), required=True)]

    @classmethod
    def create(cls, path=None):
        """
        Build a ConfigSource from the given path(s).

        Returns a YamlConfigSource if the given argument is a path and the
        file specified by that path contains no base/parents. Otherwise,
        returns a ChainedConfigSource chaining the conf to its base;
        recursively calls itself if base confs are found.

        This tends to flatten the result out a bit (removes layers of
        ChainedConfigSource), as opposed to always returning a
        ChainedConfigSource.

        FIXME: Probably errors badly if you have a cycle in your config
        (a -> b -> a). Don't do that, because we don't check for that
        currently.

        :param cls: class
        :param path: config path
        """
        # If empty list or string passed, return an conf
        return cls.load_single(path)

    @classmethod
    def load_single(cls, path):
        """
        Build a config source from a yaml file, chaining it to
        any parents that are found.

        :param cls: class
        :param path: path to yaml file
        """
        if not path:
            return ChainedConfigSource()

        conf = cls(path)
        base_conf = cls.load_base(conf.get_base(), os.path.dirname(path))
        if len(base_conf) > 0:
            return ChainedConfigSource(conf, base_conf)
        else:
            return conf

    @classmethod
    def load_base(cls, base, dirpath):
        """
        Load base configuration.

        :param cls: class
        :param base: iterable of yaml paths, possibly nested
        :param dirpath: os.path.dirname() of the child
        """

        if not base:
            return ChainedConfigSource()
        elif isinstance(base, string_type):
            return cls.load_single(os.path.join(dirpath, base))

        # If a list of things is passed, it's an optionally-nested list of
        # paths. Chain all their confs together.
        # Note: Allowing this syntax complicates things and is annoying.
        elif isinstance(base, collections.Iterable):
            confs = [cls.load_base(b, dirpath) for b in base]
            if len(confs) == 1:
                return confs[0]
            else:
                return ChainedConfigSource(*confs)
        else:
            raise Exception('error: {}', base)

    def __getitem__(self, key):
        return self.settings[key]

    def __len__(self):
        return len(self.settings)

    def __iter__(self):
        return iter(self.settings)

    def __repr__(self):
        return 'YamlConfigSource({0})'.format(self.abspath)


class CliConfigSource(ConfigSourceBase):
    """
    Configuration from the command line.
    """
    def __init__(self, overrides):
        self._overrides = {}
        for arg in overrides or []:
            k, v = arg.split('=', 1)
            self._overrides[k] = v

    def __iter__(self):
        return iter(self._overrides)

    def __len__(self):
        return len(self._overrides)

    def __getitem__(self, key):
        return self._overrides[key]



#!/usr/bin/env python

from setuptools import setup

# see https://github.com/pypa/pypi-legacy/issues/148
try:
    import pypandoc
    readme = pypandoc.convert('README.md', 'rst')
except ImportError:
    readme = open('README.md').read()


packages = [
    'interpolatr'
]

requirements = [
    'click>=6.0',
    'Jinja2>=2.8',
    'PyYAML>=3.12'
]

extras_require = {
    ":python_version < '3.0'": ['chainmap>=1.0.2']
}

with open('interpolatr/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            parts = line.split('=')
            version = parts[1].strip().strip("'")
            break

if not version:
    raise RuntimeError('Cannot find version information')

setup(
    name='interpolatr',
    version=version,
    description='Interpolation tool',
    long_description=readme,
    author='Jacob Tolar',
    author_email='development@sheckel.net',
    url='http://github.com/jacobtolar/interpolatr',
    packages=['interpolatr'],
    package_data={'': ['LICENSE']},
    include_package_data=True,
    install_requires=requirements,
    license='Apache 2.0',
    zip_safe=False,
    classifiers=(
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7'
    ),
    scripts=['scripts/interpolatr'],
    extras_require=extras_require
)



# Copyright 2017, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0.
# See the LICENSE file associated with the project for terms.

"""
Utility functions for interpolatr
"""

import interpolatr
import importlib
import click
import logging


def setup_logging(debug, log_file):
    """
    Set up logging - info or debug to console and debug to a file
    """
    fmt = '%(asctime)s - %(name)s - %(levelname)8s: %(message)s'
    level = logging.DEBUG if debug else logging.INFO

    rootLogger = logging.getLogger()
    rootLogger.setLevel(logging.NOTSET)

    streamhandler = logging.StreamHandler()
    streamhandler.setLevel(level)
    streamhandler.setFormatter(logging.Formatter(fmt))
    rootLogger.addHandler(streamhandler)

    if log_file:
        filehandler = logging.FileHandler(log_file)
        filehandler.setLevel(logging.NOTSET)
        filehandler.setFormatter(logging.Formatter(fmt))
        rootLogger.addHandler(filehandler)

import_cache = {
    'interpolatr': interpolatr
}


def lookup_import(name):
    """
    Given a name, try to find it as an interpolatr builtin, than as a module
    import.

    :param name: qualified class name
    """
    if hasattr(import_cache['interpolatr'], name):
        return getattr(import_cache['interpolatr'], name)

    dotted = name.split('.')
    classname = dotted[-1]

    if len(dotted) == 1:
        raise Exception('Module import required; specified: {0}'.format(name))
    else:
        module = '.'.join(dotted[0:-1])

    if module not in import_cache:
        import_cache[module] = importlib.import_module(module)

    return getattr(import_cache[module], classname)


class InterpolatrCommand(click.MultiCommand):
    """
    Command-line class that allows dynamic lookup of commands.
    """
    def get_command(self, ctx, name):
        """
        Given a command name, build the command.

        :param ctx: click context.
        :param name: command name
        """
        cls = lookup_import(name)
        args = cls.get_args()

        @click.pass_context
        def callback(ctx, *args, **kwargs):
            return cls.create(*args, **kwargs)

        return click.Command(name, params=args, callback=callback)

    def format_usage(self, ctx, formatter):
        """
        Usage for the interpolatr command.

        Overrides the default usage to make the expected usage more clear:
        the command should specify a conf or provider class.
        """
        pieces = [
            '[OPTIONS]',
            '[<conf-class> [ARGS]...]...',
            '[<supplier-class> [ARGS]...]...'
        ]

        formatter.write_usage(ctx.command_path, ' '.join(pieces))
